from __future__ import annotations

from troposphere import Ref
from e3.aws.troposphere import Stack
from e3.aws.troposphere.cloudwatch import Alarm


EXPECTED_ALARM_DEFAULT_TEMPLATE = {
    "Myalarm": {
        "Properties": {
            "AlarmDescription": "Some description",
            "AlarmName": "myalarm",
            "ComparisonOperator": "LessThanThreshold",
            "EvaluationPeriods": 2,
            "Statistic": "Sum",
        },
        "Type": "AWS::CloudWatch::Alarm",
    },
}


EXPECTED_ALARM_TEMPLATE = {
    "Myalarm": {
        "Properties": {
            **EXPECTED_ALARM_DEFAULT_TEMPLATE["Myalarm"]["Properties"],
            **{
                "AlarmActions": ["StrAction", {"Ref": "RefAction"}],
                "Dimensions": [
                    {"Name": "StrDimensionValue", "Value": "DimensionValue"},
                    {"Name": "RefDimensionValue", "Value": {"Ref": "DimensionValue"}},
                ],
                "MetricName": "Invocations",
                "Namespace": "AWS/Lambda",
                "Period": 300,
                "Threshold": 1.0,
            },
        },
        "Type": "AWS::CloudWatch::Alarm",
    },
}


def test_alarm_default(stack: Stack) -> None:
    """Test default Alarm creation."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"
    stack.add(
        Alarm(
            name="myalarm",
            description="Some description",
            comparison_operator="LessThanThreshold",
            evaluation_periods=2,
            statistic="Sum",
        )
    )
    assert stack.export()["Resources"] == EXPECTED_ALARM_DEFAULT_TEMPLATE


def test_alarm(stack: Stack) -> None:
    """Test Alarm creation."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"
    stack.add(
        Alarm(
            name="myalarm",
            description="Some description",
            comparison_operator="LessThanThreshold",
            evaluation_periods=2,
            actions=["StrAction", Ref("RefAction")],
            dimensions={
                "StrDimensionValue": "DimensionValue",
                "RefDimensionValue": Ref("DimensionValue"),
            },
            metric_name="Invocations",
            namespace="AWS/Lambda",
            period=300,
            statistic="Sum",
            threshold=1.0,
        )
    )
    assert stack.export()["Resources"] == EXPECTED_ALARM_TEMPLATE



"""Provide Cloudformation construct tests."""

from __future__ import annotations

from troposphere import AccountId, Equals

from e3.aws.troposphere import Stack
from e3.aws.troposphere.cloudformation import StackSet
from e3.aws.troposphere.iam.role import Role


EXPECTED_TEMPLATE = {
    "StackSetTest": {
        "Properties": {
            "AutoDeployment": {"Enabled": True, "RetainStacksOnAccountRemoval": False},
            "CallAs": "SELF",
            "Capabilities": ["CAPABILITY_NAMED_IAM"],
            "Description": "this is a test",
            "PermissionModel": "SERVICE_MANAGED",
            "StackSetName": "stack-set-test",
            "TemplateURL": "https://cfn_bucket.s3.amazonaws.com/templates/"
            "stack-set-test-template.yaml",
            "OperationPreferences": {
                "FailureToleranceCount": 4,
                "MaxConcurrentCount": 5,
            },
            "StackInstancesGroup": [
                {
                    "DeploymentTargets": {"OrganizationalUnitIds": ["test-ou"]},
                    "Regions": ["eu-west-1"],
                }
            ],
        },
        "Type": "AWS::CloudFormation::StackSet",
    }
}


def test_stackset(stack: Stack) -> None:
    """test Cloudformation stack set creation."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"

    stack_set = StackSet(
        name="stack-set-test",
        description="this is a test",
        regions=["eu-west-1"],
        ous=["test-ou"],
        failure_tolerance_count=4,
        max_concurrent_count=5,
    )

    stack_set.add(
        Role(
            name="TestRole",
            description="TestRole description",
            trust={"Service": "test"},
        )
    )
    stack_set.add_condition("", Equals(AccountId, "test_account_id"))
    stack.add(stack_set)
    assert stack.export()["Resources"] == EXPECTED_TEMPLATE



from e3.aws.cfn import AWSType, Resource


class RecordSet(Resource):
    """DNS Record."""

    def __init__(self, name, hosted_zone, dns_name, dns_type, ttl, resource_records):
        """Initialize a DNS Record.

        :param name: logical name used in the stack
        :type name: str
        :param hosted_zone: name of the domain for the hosted zone
            or HostedZone object
        :type hosted_zone: str | HostedZone
        :param dns_name: domain name (fqdn)
        :type dns_name: str
        :param dns_type: record type
        :type dns_type: str
        :param ttl: dns TTL
        :type ttl: int
        :param resource_records: list of resourses associated with the entry
        :type resource_records: list[str]
        """
        super(RecordSet, self).__init__(name, kind=AWSType.ROUTE53_RECORDSET)
        self.hosted_zone = hosted_zone
        self.dns_name = dns_name
        self.dns_type = dns_type
        self.ttl = ttl
        self.resource_records = resource_records

    @property
    def properties(self):
        result = {
            "Name": self.dns_name,
            "Type": self.dns_type,
            "TTL": self.ttl,
            "ResourceRecords": self.resource_records,
        }
        if isinstance(self.hosted_zone, HostedZone):
            result["HostedZoneId"] = self.hosted_zone.ref
        else:
            result["HostedZoneName"] = self.hosted_zone
        return result


class HostedZone(Resource):
    """Hosted Zone."""

    def __init__(self, name, domain, vpcs=None):
        """Initialize an hosted zone.

        :param name: logical name in the stack
        :type name: str
        :param domain: domain name (end it with a dot to make it absolute)
        :type domain: str
        :param vpcs: list of vpcs associated with the domain. If the list
            is empty it means that a public zone is created. Otherwise
            this is a AWS private zone.
        :type vpcs: None | list[VPC]
        """
        super(HostedZone, self).__init__(name, kind=AWSType.ROUTE53_HOSTED_ZONE)
        self.domain = domain
        self.vpcs = vpcs

    @property
    def properties(self):
        result = {"Name": self.domain}
        if self.vpcs is not None:
            result["VPCs"] = [
                {"VPCId": vpc.ref, "VPCRegion": vpc.region} for vpc in self.vpcs
            ]
        return result



from setuptools import setup, find_packages
import os

# Get e3 version from the VERSION file.
version_file = os.path.join(os.path.dirname(__file__), "VERSION")
with open(version_file) as f:
    e3_version = f.read().strip()

setup(
    name="e3-aws",
    version=e3_version,
    description="E3 Cloud Formation Extension",
    author="AdaCore's Production Team",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=(
        "boto3",
        "botocore",
        # Only require docker for linux as there is a known dependency issue
        # with pywin32 on windows
        "docker; platform_system=='Linux'",
        "pyyaml",
        "troposphere",
        "e3-core",
    ),
    namespace_packages=["e3"],
    entry_points={
        "e3.event.handler": ["s3-boto3 = e3.aws.handler.s3:S3Handler"],
        "console_scripts": [
            "e3-aws-assume-profile = e3.aws:assume_profile_main",
            "e3-aws-assume-role = e3.aws:assume_role_main",
        ],
    },
)



from __future__ import annotations
from dataclasses import dataclass, field
from typing import TYPE_CHECKING


from troposphere import AWSObject, ecs, GetAtt, Ref, Tags

from e3.aws import name_to_id
from e3.aws.troposphere import Construct
from e3.aws.troposphere.iam.role import Role
from e3.aws.troposphere.iam.managed_policy import ManagedPolicy
from e3.aws.troposphere.iam.policy_statement import PolicyStatement, Trust

if TYPE_CHECKING:
    from e3.aws.troposphere import Stack


@dataclass(frozen=True)
class Cluster(Construct):
    """Define a Cluster construct.

    :param name: a string that you use to identify your cluster
    :param cluster_settings: The setting to use when creating a cluster. This
        parameter is used to enable CloudWatch Container Insights for a cluster
    :param default_capacity_provider_strategy: The default capacity provider strategy
        for the cluster
    :param tags: The metadata that you apply to the cluster to help you categorize and
        organize them
    """

    name: str
    capacity_providers: list[str] | None = None
    cluster_settings: list[dict[str, str]] | None = None
    default_capacity_provider_strategy: list[dict[str, str]] | None = None
    tags: dict[str, str] = field(default_factory=lambda: {})

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Construct and return ECS cluster troposphere resources."""
        c_settings = None
        if self.cluster_settings:
            c_settings = [ecs.ClusterSetting(**cs) for cs in self.cluster_settings]

        provider_strategy = None
        if self.default_capacity_provider_strategy:
            provider_strategy = [
                ecs.CapacityProviderStrategyItem(**ps)
                for ps in self.default_capacity_provider_strategy
            ]

        kwargs = {
            key: val
            for key, val in {
                "ClusterName": self.name,
                "ClusterSettings": c_settings,
                "CapacityProviders": self.capacity_providers,
                "DefaultCapacityProviderStrategy": provider_strategy,
                "Tags": Tags({"Name": self.name, **self.tags}),
            }.items()
            if val is not None
        }

        return [ecs.Cluster(name_to_id(self.name), **kwargs)]


class ECSPassExecutionRolePolicy(ManagedPolicy):
    """ECSPassExecutionRolePolicy, see description for details."""

    def __init__(self, path: str) -> None:
        """Initialize ECSPassExecutationRolePolicy.

        :param path: path for the managed policy
        """
        super().__init__(
            name="ECSPassExecutionRolePolicy",
            description="Needed to be attached to ECSEventsRole if schedulded"
            "task requires ECSTaskExecutionRole",
            statements=[
                PolicyStatement(
                    effect="Allow",
                    action=["iam:PassRole"],
                    resource=GetAtt(name_to_id("ECSTaskExecutionRole"), "Arn"),
                )
            ],
            path=path,
        )


@dataclass(frozen=True)
class FargateCluster(Cluster):
    """Define a FargateCluster construct.

    :param name: a string that you use to identify your cluster
    :param path: path for IAM resources
    :param tags: The metadata that you apply to the cluster to help you categorize and
        organize them
    """

    name: str
    path: str = "/"
    capacity_providers: list[str] | None = field(default_factory=lambda: ["FARGATE"])
    cluster_settings: list[dict[str, str]] | None = field(
        default_factory=lambda: [{"Name": "containerInsights", "Value": "enabled"}]
    )
    default_capacity_provider_strategy: list[dict[str, str]] | None = field(
        default_factory=lambda: [{"CapacityProvider": "FARGATE", "Weight": "1"}]
    )
    tags: dict[str, str] = field(default_factory=lambda: {})

    @property
    def ecs_task_execution_role(self) -> Role:
        """Return ecs task execution role, see role description for details."""
        # For unknown reason if path is provided for this role it cannot be
        # use by ECS task definitions.
        return Role(
            name="ECSTaskExecutionRole",
            description="grants the Amazon ECS container agent permission to make "
            "AWS API calls on your behalf.",
            trust=Trust(services=["ecs-tasks"]),
            managed_policy_arns=[
                "arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy"
            ],
        )

    @property
    def ecs_events_role(self) -> Role:
        """Return ecs events role, see role description for details."""
        return Role(
            name="ECSEventsRole",
            description="Allow CloudWatch Events service to run Amazon ECS tasks",
            trust=Trust(services=["events"]),
            managed_policy_arns=[
                "arn:aws:iam::aws:policy/service-role/"
                "AmazonEC2ContainerServiceEventsRole",
                Ref(name_to_id("ECSPassExecutionRolePolicy")),
            ],
            path=self.path,
        )

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Construct and return Fargate ECS cluster and associated resources.

        An IAM role for Fargates tasks to be used is also returned.
        """
        return (
            super().resources(stack=stack)
            + ECSPassExecutionRolePolicy(path=self.path).resources(stack=stack)
            + self.ecs_task_execution_role.resources(stack=stack)
            + self.ecs_events_role.resources(stack=stack)
        )



"""Provide Cloudformation construct tests."""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from e3.aws import AWSEnv
from e3.aws.troposphere import CFNProjectMain
from e3.aws.troposphere.iam.role import Role


if TYPE_CHECKING:
    from e3.aws.troposphere import Stack


TEST_DIR = os.path.dirname(os.path.abspath(__file__))


class MyCFNProject(CFNProjectMain):
    """Provide CLI to manage MyCFNProject."""

    def create_stack(self) -> list[Stack]:
        """Return MyCFNProject stack."""
        self.add(
            (
                Role(
                    name="TestRole",
                    description="TestRole description",
                    trust={"Service": "test"},
                )
            )
        )
        return self.stack


def test_cfn_project_main(capfd) -> None:
    """Test CFNProjectMain."""
    aws_env = AWSEnv(regions=["eu-west-1"], stub=True)
    test = MyCFNProject(
        name="TestProject",
        account_id="12345678",
        stack_description="TestStack",
        s3_bucket="cfn-test-deploy-bucket",
        regions=["eu-west-1"],
    )
    test.execute(args=["show"], aws_env=aws_env)

    captured = capfd.readouterr()
    print(captured.out)
    with open(os.path.join(TEST_DIR, "cfn_project_test.out")) as f_out:
        assert captured.out == f_out.read()



from __future__ import annotations
import pytest

from e3.aws import AWSEnv, AWSSessionRunError
from e3.aws import Session
from datetime import datetime

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _pytest.capture import CaptureFixture


def test_run(capfd: CaptureFixture) -> None:
    """Test AWS session run method."""
    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    stubber = aws_env.stub("sts")

    # 2 calls to cli_cmd are made in this test
    for _ in range(2):
        stubber.add_response(
            "assume_role",
            {
                "Credentials": {
                    "AccessKeyId": "12345678912345678",
                    "SecretAccessKey": "12345678912345678",
                    "SessionToken": "12345678912345678",
                    "Expiration": datetime(4042, 1, 1),
                }
            },
            {
                "RoleArn": "arn:aws:iam::123456789123:role/TestRole",
                "RoleSessionName": "aws_run_session",
                "DurationSeconds": 7200,
            },
        )

    with stubber:
        p_right = aws_env.run(
            ["aws", "--version"],
            "arn:aws:iam::123456789123:role/TestRole",
            output=None,
            session_duration=7200,
        )
        assert p_right.status == 0
        captured = capfd.readouterr()
        assert captured.out.startswith("aws-cli/")

        with pytest.raises(AWSSessionRunError):
            p_wrong = aws_env.run(
                ["aws", "not_a_command"],
                "arn:aws:iam::123456789123:role/TestRole",
                output=None,
                session_duration=7200,
            )
            assert p_wrong.status != 0


def test_boto3() -> None:
    """Test boto3 session from Session."""
    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    stubber = aws_env.stub("sts")

    with stubber:
        session = Session(
            regions=["us-east-1"],
            credentials={
                "AccessKeyId": "AK_test",
                "SecretAccessKey": "SA_test",
                "SessionToken": "ST_test",
            },
        )
        boto3_session = session.to_boto3()
        boto3_creds = boto3_session.get_credentials().get_frozen_credentials()

        assert boto3_creds.access_key == "AK_test"
        assert boto3_creds.secret_key == "SA_test"
        assert boto3_creds.token == "ST_test"



from e3.aws.cfn.ec2 import VPC
from e3.aws.cfn.service_discovery import PrivateDnsNamespace


def test_create_privatedns():
    vpc = VPC("myvpc", "10.10.0.0/16")
    p = PrivateDnsNamespace("mypriv", vpc, domain="vpc.", description="Priv Namespace")
    assert p.properties



"""Utility functions for AWS ECR service."""

from __future__ import annotations

import base64
import logging
import tempfile
from typing import TYPE_CHECKING

import docker
from e3.fs import sync_tree, rm

if TYPE_CHECKING:
    from e3.aws import Session

logger = logging.getLogger("e3.aws.utils.ecr")


def build_and_push_image(
    source_dir: str, repository_name: str, image_tag: str, session: Session
) -> str:
    """Build and push image to an ECR repository and return image URI.

    :param source_dir: directory where to find Dockerfile
    :param repository_name: ECR repository name
    :param image_tag: Docker image tag
    :param session: AWS session to push docker image to ECR
    """
    temp_dir = tempfile.mkdtemp()
    try:
        sync_tree(source_dir, temp_dir)

        docker_client = docker.from_env()
        # Build image
        image, build_log = docker_client.images.build(
            path=temp_dir, tag=f"{repository_name}:{image_tag}", rm=True
        )
        logging.info(f"Building Docker image from {source_dir}")
        for chunk in build_log:
            if "stream" in chunk:
                for line in chunk["stream"].splitlines():
                    logging.debug(line)

        # Push image to registry
        ecr_client = session.client("ecr")
        ecr_creds = ecr_client.get_authorization_token()["authorizationData"][0]
        ecr_username, ecr_password = (
            base64.b64decode(ecr_creds["authorizationToken"])
            .decode("utf-8")
            .split(":", 1)
        )
        ecr_url = ecr_creds["proxyEndpoint"]
        docker_client.login(
            username=ecr_username, password=ecr_password, registry=ecr_url
        )
        ecr_repo_name = f"{ecr_url.replace('https://', '')}/{repository_name}"
        image.tag(ecr_repo_name, tag=image_tag)

        push_log = docker_client.images.push(ecr_repo_name, tag=image_tag)
        logging.info(f"Pushing Docker image to {ecr_repo_name} with tag {image_tag}")
        logging.debug(push_log)

        image_uri = f"{ecr_repo_name}:{image_tag}"

    finally:
        rm(temp_dir, recursive=True)

    return image_uri



from __future__ import annotations
from typing import TYPE_CHECKING
from troposphere import cloudwatch, GetAtt, Ref

from e3.aws import name_to_id
from e3.aws.troposphere import Construct

if TYPE_CHECKING:
    from troposphere import AWSObject
    from e3.aws.troposphere import Stack


class Alarm(Construct):
    """A CloudWatch Alarm."""

    def __init__(
        self,
        name: str,
        description: str,
        comparison_operator: str,
        evaluation_periods: int,
        actions: list[str | Ref] | None = None,
        dimensions: dict[str, str | Ref] | None = None,
        metric_name: str | None = None,
        namespace: str | None = None,
        period: int | None = None,
        statistic: str | None = None,
        threshold: float | None = None,
    ):
        """Initialize an AWS CloudWatch alarm.

        :param name: alarm name
        :param description: the description of the alarm
        :param comparison_operator: the arithmetic operation to use when
            comparing the specified statistic and threshold
        :param evaluation_periods: the number of periods over which data is
            compared to the specified threshold
        :param actions: the list of actions to execute when this alarm transitions
            into an ALARM state from any other state
        :param dimensions: the dimensions for the metric associated with the alarm
        :param metric_name: the name of the metric associated with the alarm
        :param namespace: the namespace of the metric associated with the alarm
        :param period: the period, in seconds, over which the statistic is applied
        :param statistic: the statistic for the metric associated with the alarm,
            other than percentile. For percentile statistics, use ExtendedStatistic
        :param threshold: the value to compare with the specified statistic
        """
        self.name = name
        self.description = description
        self.comparison_operator = comparison_operator
        self.evaluation_periods = evaluation_periods
        self.actions = actions
        self.dimensions = dimensions
        self.metric_name = metric_name
        self.namespace = namespace
        self.period = period
        self.statistic = statistic
        self.threshold = threshold

    @property
    def arn(self) -> GetAtt:
        """Arn of the CloudWatch alarm."""
        return GetAtt(name_to_id(self.name), "Arn")

    @property
    def ref(self) -> Ref:
        return Ref(name_to_id(self.name))

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return list of AWSObject associated with the construct."""
        params = {
            "AlarmName": self.name,
            "AlarmDescription": self.description,
            "ComparisonOperator": self.comparison_operator,
            "EvaluationPeriods": self.evaluation_periods,
        }

        if self.actions is not None:
            params["AlarmActions"] = self.actions

        if self.dimensions is not None:
            params["Dimensions"] = [
                cloudwatch.MetricDimension(Name=k, Value=v)
                for k, v in self.dimensions.items()
            ]

        if self.metric_name is not None:
            params["MetricName"] = self.metric_name

        if self.namespace is not None:
            params["Namespace"] = self.namespace

        if self.period is not None:
            params["Period"] = self.period

        if self.statistic is not None:
            params["Statistic"] = self.statistic

        if self.threshold is not None:
            params["Threshold"] = self.threshold

        return [cloudwatch.Alarm(name_to_id(self.name), **params)]



from __future__ import annotations
from enum import Enum
from typing import TYPE_CHECKING

from e3.aws.cfn import AWSType, GetAtt, Resource, Stack

if TYPE_CHECKING:
    from typing import Any, Iterable


class PrincipalKind(Enum):
    AWS = "AWS"
    FEDERATED = "Federated"
    SERVICE = "Service"
    EVERYONE = "*"


class Principal(object):
    """Represent a principal in an IAM policy."""

    def __init__(self, kind: PrincipalKind, value: str | None = None) -> None:
        """Initialize a Principal.

        :param kind: principal kind
        :param value: string value for the principal. If kind is set
            to EVERYONE value should be None, otherwise value should
            be a string different from '*'
        """
        assert isinstance(kind, PrincipalKind)
        self.kind = kind
        assert (kind == PrincipalKind.EVERYONE and value is None) or (
            value is not None and value != "*"
        )
        self.value = value

    @classmethod
    def property_list(cls, principals):
        """Serialize a list of principal as a simple object.

        :param principals: list of principals
        :type principals: list[Principal]
        :rtype: dict | str
        """
        result = {}
        for principal in principals:
            if principal.kind == PrincipalKind.EVERYONE:
                # If EVERYONE is present then it should be alone because
                # it will mask any other principal.
                assert len(principals) == 1, 'Principal "*" should be used alone'
                result = "*"
                break

            if principal.kind.value not in result:
                result[principal.kind.value] = []
            result[principal.kind.value].append(principal.value)

        return result


class Statement(object):
    """Statement of IAM Policy Document."""

    def __init__(
        self,
        sid: str | None = None,
        to: list[str] | str | None = None,
        on: list[str] | str | None = None,
        not_on: list[str] | str | None = None,
        apply_to: list[Principal] | Principal | None = None,
    ):
        """Initialize a statement.

        :param sid: statement id (optional)
        :param to: one or several action for the statement
        :param on: resource or list of resources on which the statement apply
        :type on: list[str] | str | None
        :param not_on: resource or list of resources on which the statement
            does not apply. Note that not_on and on cannot be both set
        :type not_on: list[str] | str | None
        :param apply_to: list of principals that are targeted by the statement
        :type apply_to: list[Principals] | Principal | None
        """
        self.resources: list[Any] = []
        self.not_resources: list[Any] = []
        self.actions: list[Any] = []
        self.principals: list[Any] = []
        self.sid = sid
        self.condition = None

        if to is not None:
            self.to(to)
        if on is not None:
            self.on(on)
        if not_on is not None:
            self.not_on(not_on)
        if apply_to is not None:
            self.apply_to(apply_to)

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        result = {"Effect": self.EFFECT}
        if self.sid is not None:
            result["Sid"] = self.sid
        if self.resources:
            result["Resource"] = self.resources
        if self.not_resources:
            result["NotResource"] = self.not_resources
        if self.principals:
            result["Principal"] = Principal.property_list(self.principals)
        if self.actions:
            result["Action"] = self.actions
        if self.condition is not None:
            result["Condition"] = self.condition
        return result

    def to(self, actions: Iterable[str] | str | GetAtt) -> Statement:
        """Add action(s) targeted by statement.

        :param actions: one or several action for the statement
        :return: the modified statement
        """
        if isinstance(actions, str) or isinstance(actions, GetAtt):
            self.actions.append(actions)
        else:
            self.actions += actions
        return self

    def on(self, resources: Iterable[str | GetAtt] | str | GetAtt) -> Statement:
        """Add resource(s) on which the statement apply.

        :param resources: resource or list of resources
        :return: the modified statement
        """
        assert not self.not_resources
        if isinstance(resources, str) or isinstance(resources, GetAtt):
            resources = [resources]
        self.resources += resources
        return self

    def not_on(self, resources: Iterable[str | GetAtt] | str | GetAtt) -> Statement:
        """Add resource(s) on which not to apply the statement.

        :param resources: resource or list of resources
        :return: the modified statement
        """
        assert not self.resources
        if isinstance(resources, str) or isinstance(resources, GetAtt):
            resources = [resources]
        self.not_resources += resources
        return self

    def apply_to(self, principals: list[Principal] | Principal) -> Statement:
        """Add principal that can use the statement.

        :param principals: principal list
        :type principals: list[Principal] | Principal
        :return: the modified statement
        :rtype: Statement
        """
        if isinstance(principals, Principal):
            self.principals.append(principals)
        else:
            self.principals += principals
        return self


class Allow(Statement):
    """Allow statement."""

    EFFECT = "Allow"


class Deny(Statement):
    """Deny statement."""

    EFFECT = "Deny"


class PolicyDocument:
    """IAM Policy Document."""

    def __init__(self):
        """Initialize a policy document."""
        self.statements = []

    def append(self, statement):
        """Append a statement.

        :param statement: a IAM Statement
        :type statement: Statement
        :return: the modified policy document
        :rtype: PolicyDocument
        """
        assert isinstance(statement, Statement)
        self.statements.append(statement)
        return self

    def extend(self, statements):
        """Append a list of statements.

        :param statements: IAM Statements
        :type statements: list[Statement]
        :return: the modified policy document
        :rtype: PolicyDocument
        """
        for s in statements:
            assert isinstance(s, Statement)
        self.statements.extend(statements)
        return self

    def __iadd__(self, statements):
        """see extend."""
        return self.extend(statements)

    def __add__(self, other):
        """Add statement(s) or merge two policy documents.

        :param other: statement, list of statements or policy document
        :type other: Statement | list[Statement] | PolicyDocument
        :return: the modified policy document
        :rtype: PolicyDocument
        """
        result = PolicyDocument()
        result.extend(self.statements)
        if isinstance(other, Statement):
            result.append(other)
        elif isinstance(other, PolicyDocument):
            result.extend(other.statements)
        else:
            result.extend(other)
        return result

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        assert self.statements, "A policy should have at least one statement"
        return {
            "Version": "2012-10-17",
            "Statement": [s.properties for s in self.statements],
        }


INSTANCE_ASSUME_ROLE = Allow(
    to="sts:AssumeRole", apply_to=Principal(PrincipalKind.SERVICE, "ec2.amazonaws.com")
)


class Policy(Resource):
    """A CloudFormation Policy resource."""

    def __init__(self, name, policy_document=None, roles=None, groups=None, users=None):
        """Initialize a policy.

        :param name: logical name on the stack
        :type name: str
        :param policy_document: policy document
        :type policy_document: PolicyDocument
        :param roles: list of roles to apply the policy to
        :type roles: list[str] | None
        :param groups: list of groups to apply the policy to
        :type groups: list[str] | None
        :param users: list of users to apply the policy to
        :type users: list[str] | None
        """
        super(Policy, self).__init__(name, kind=AWSType.IAM_POLICY)
        self.roles = roles
        self.groups = groups
        self.users = users
        self.policy_document = policy_document

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        result = {"PolicyName": self.name}
        if self.policy_document is not None:
            result["PolicyDocument"] = self.policy_document.properties

        if self.roles is not None:
            result["Roles"] = self.roles
        if self.groups is not None:
            result["Groups"] = self.groups
        if self.users is not None:
            result["Users"] = self.users
        return result


class InstanceProfile(Resource):
    """IAM Instance profile."""

    def __init__(self, name, role):
        """Initialize an instance profile.

        :param name: logical name in the stack
        :type name: str
        :param role: name of the associated role
        :type role: str
        """
        super(InstanceProfile, self).__init__(name, kind=AWSType.IAM_INSTANCE_PROFILE)
        self.role = role

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        return {"Roles": [self.role], "Path": "/"}


class Role(Resource):
    """IAM Role."""

    def __init__(self, name, assume_role_policy, path="/"):
        """Initialize IAM Role.

        :param name: role name
        :type name: str
        :param assume_role_policy: policy to define who can assume the role
        :type assume_role_policy: PolicyDocument
        :param path: the path associated with this role (default: /)
        :type path: str
        """
        # Note: we don't set RoleName attribute because of limitations during
        # update with cloudform. In that case RoleName is generated directly by
        # Cloud Formation. This is not related to the name of the resource as
        # part of a stack.
        super(Role, self).__init__(name, kind=AWSType.IAM_ROLE)
        self.path = path
        self.policies = []
        self.assume_role_policy = assume_role_policy

    def add(self, policy):
        """Add a policy.

        :param policy: a policy
        :type policy: Policy
        :return: the object itself
        :rtype: Role
        """
        assert isinstance(policy, Policy)
        self.policies.append(policy)
        return self

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        return {
            "AssumeRolePolicyDocument": self.assume_role_policy.properties,
            "Path": self.path,
            "Policies": [p.properties for p in self.policies],
        }


class Group(Resource):
    """IAM Group."""

    ATTRIBUTES = ("Arn",)

    def __init__(self, name, managed_policy_arns=None, path="/"):
        """Initialize IAM Group.

        :param name: group name
        :type name: str
        :param managed_policy_arns: A list of Amazon Resource Names (ARNs) of
            the IAM managed policies that you want to attach to the user.
        :type managed_policy_arns: list[str] | None
        :param path: the path associated with this role (default: /)
        :type path: str
        """
        super(Group, self).__init__(name, kind=AWSType.IAM_GROUP)
        self.path = path
        self.policies = []
        self.managed_policy_arns = managed_policy_arns or []

    def add(self, policy):
        """Add a policy.

        :param policy: a policy
        :type policy: Policy
        :return: the object itself
        :rtype: Role
        """
        assert isinstance(policy, Policy)
        self.policies.append(policy)
        return self

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        return {
            "ManagedPolicyArns": self.managed_policy_arns,
            "GroupName": self.name,
            "Path": self.path,
            "Policies": [p.properties for p in self.policies],
        }


class User(Resource):
    """IAM User."""

    ATTRIBUTES = ("Arn",)

    def __init__(
        self,
        name,
        groups=None,
        managed_policy_arns=None,
        path="/",
        permissions_boundary=None,
    ):
        """Initialize IAM User.

        :param name: user name
        :type name: str
        :type groups: list[str] | None
        :param managed_policy_arns: A list of Amazon Resource Names (ARNs) of
            the IAM managed policies that you want to attach to the user.
        :type managed_policy_arns: list[str] | None
        :param path: the path associated with this role (default: /)
        :type path: str
        :param permissions_boundary: The ARN of the policy that is used to set
            the permissions boundary for the user.
        :type permissions_boundary: str | None
        """
        super(User, self).__init__(name, kind=AWSType.IAM_USER)
        self.groups = groups or []
        self.path = path
        self.policies = []
        self.managed_policy_arns = managed_policy_arns or []
        self.permissions_boundary = permissions_boundary

    def add(self, policy):
        """Add a policy.

        :param policy: a policy
        :type policy: Policy
        :return: the object itself
        :rtype: Role
        """
        assert isinstance(policy, Policy)
        self.policies.append(policy)
        return self

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        props = {
            "ManagedPolicyArns": self.managed_policy_arns or [],
            "UserName": self.name,
            "Groups": self.groups,
            "Path": self.path,
            "Policies": [p.properties for p in self.policies],
        }
        if self.permissions_boundary is not None:
            props["PermissionsBoundary"] = self.permissions_boundary
        return props


class InstanceRole(Stack):
    """Instance Role.

    Create both Role and associated profile.
    """

    def __init__(self, name, path="/"):
        """Initialize an instance role.

        :param name: name of the role. The instance profile name will be this
            name + ``InstanceProfile``
        :type name: str
        :param path: path associated with this role
        :type path: str
        """
        super(InstanceRole, self).__init__(name)
        assume_role_policy = PolicyDocument()
        assume_role_policy.append(INSTANCE_ASSUME_ROLE)
        self.add(Role(name, assume_role_policy, path))
        self.add(InstanceProfile(name + "InstanceProfile", self.resources[name].ref))

    def add_policy(self, policy):
        """Add a policy.

        :param policy: a policy
        :type policy: Policy
        :return: the object itself
        :rtype: InstanceRole
        """
        self[self.name].add(policy)
        return self

    @property
    def instance_profile(self):
        """Return the name of the instance profile.

        :return: the name of the instance profile
        :rtype: str
        """
        return self.resources[self.name + "InstanceProfile"]



from __future__ import annotations
import os
from typing import TYPE_CHECKING


from troposphere import AccountId, cloudfront, GetAtt, Join, route53, Ref, Sub

from e3.aws import name_to_id
from e3.aws.troposphere import Construct
from e3.aws.troposphere.awslambda import Function
from e3.aws.troposphere.iam.managed_policy import ManagedPolicy
from e3.aws.troposphere.iam.policy_statement import Allow, Trust
from e3.aws.troposphere.iam.role import Role
from e3.aws.troposphere.s3.bucket import Bucket
from e3.aws.troposphere.sns import Topic

if TYPE_CHECKING:
    from typing import Any
    from troposphere import AWSObject
    from e3.aws.troposphere import Stack


class S3WebsiteDistribution(Construct):
    """Set a Cloudfront distribution in front of a website hosted in S3.

    It also provides a lambda invalidating cloudfront cache when s3 objects
    are updated.
    """

    def __init__(
        self,
        name: str,
        aliases: list[str],
        bucket_name: str,
        certificate_arn: str,
        default_ttl: int = 86400,
        lambda_edge_function_arns: list[str] | None = None,
        root_object: str = "index.html",
        r53_route_from: list[tuple[str, str]] | None = None,
        logging_bucket: str | None = None,
        logging_prefix: str | None = None,
        logging_include_cookies: bool | None = None,
    ):
        """Initialize a S3WebsiteCFDistribution.

        :param name: function name
        :param aliases: CNAMEs (alternate domain names), if any, for this
            distribution.
        :param bucket_name: name of the bucket to create to host the website
        :param certificate_arn: Amazon Resource Name (ARN) of the ACM
            certificate for aliases. Cloudfront only supports ceritificates
            stored in us-east-1.
        :param default_ttl: The default amount of time, in seconds, that you
            want objects to stay in the CloudFront cache before CloudFront sends
            another request to the origin to see if the object has been updated
        :param lambda_edge_function_arns: ARNs of Lambda@Edge functions to
            associate with the cloudfront distribution default cache behaviour
        :param root_object: The object that you want CloudFront to request from
            your origin
        :param r53_route_from: list of (hosted_zone_id, domain_id) for which to
            create route53 records
        :param logging_bucket: the Amazon S3 bucket to store the access logs in,
            for example, myawslogbucket.s3.amazonaws.com
        :param logging_prefix: an optional string that you want CloudFront to
            prefix to the access log filenames
        :param logging_include_cookies: specifies whether you want CloudFront
            to include cookies in access logs, specify true for IncludeCookies
        """
        self.name = name
        self.aliases = aliases
        self.bucket = Bucket(name=bucket_name)
        self.certificate_arn = certificate_arn
        self.default_ttl = default_ttl
        self.lambda_edge_function_arns = lambda_edge_function_arns
        self.root_object = root_object
        self.r53_route_from = r53_route_from
        self._origin_access_identity = None
        self._cache_policy = None
        self.logging_bucket = logging_bucket
        self.logging_prefix = logging_prefix
        self.logging_include_cookies = logging_include_cookies

    def add_oai_access_to_bucket(self) -> None:
        """Add policy granting cloudfront OAI read access to the bucket."""
        cf_principal = {
            "CanonicalUser": GetAtt(self.origin_access_identity, "S3CanonicalUserId")
        }
        self.bucket.policy_statements.extend(
            [
                Allow(
                    action="s3:GetObject",
                    resource=self.bucket.all_objects_arn,
                    principal=cf_principal,
                ),
                Allow(
                    action="s3:ListBucket",
                    resource=self.bucket.arn,
                    principal=cf_principal,
                ),
            ]
        )

    @property
    def cache_policy(self) -> cloudfront.CachePolicy:
        """Return cloudfront distribution cache policy."""
        if self._cache_policy is None:
            forwarded_to_origin = cloudfront.ParametersInCacheKeyAndForwardedToOrigin(
                CookiesConfig=cloudfront.CacheCookiesConfig(CookieBehavior="none"),
                EnableAcceptEncodingBrotli="true",
                EnableAcceptEncodingGzip="true",
                HeadersConfig=cloudfront.CacheHeadersConfig(HeaderBehavior="none"),
                QueryStringsConfig=cloudfront.CacheQueryStringsConfig(
                    QueryStringBehavior="none"
                ),
            )
            self._cache_policy = cloudfront.CachePolicy(
                name_to_id(f"{self.name}-cloudfront-cache-policy"),
                CachePolicyConfig=cloudfront.CachePolicyConfig(
                    Comment=f"{self.name} s3 website cloudfront cache policy",
                    DefaultTTL=self.default_ttl,
                    MaxTTL=31536000,
                    MinTTL=1,
                    Name="s3-cache-policy",
                    ParametersInCacheKeyAndForwardedToOrigin=forwarded_to_origin,
                ),
            )
        return self._cache_policy

    @property
    def distribution(self) -> cloudfront.Distribution:
        """Return cloudfront distribution with bucket as origin."""
        origin = cloudfront.Origin(
            S3OriginConfig=cloudfront.S3OriginConfig(
                OriginAccessIdentity=Join(
                    "",
                    [
                        "origin-access-identity/cloudfront/",
                        Ref(self.origin_access_identity),
                    ],
                )
            ),
            DomainName=f"{self.bucket.name}.s3.amazonaws.com",
            Id="S3Origin",
        )
        cache_params = {
            "AllowedMethods": ["GET", "HEAD", "OPTIONS"],
            "CachePolicyId": Ref(self.cache_policy),
            "TargetOriginId": "S3Origin",
            "ViewerProtocolPolicy": "redirect-to-https",
        }
        if self.lambda_edge_function_arns:
            cache_params["LambdaFunctionAssociations"] = [
                cloudfront.LambdaFunctionAssociation(
                    EventType="viewer-request", LambdaFunctionARN=lambda_arn
                )
                for lambda_arn in self.lambda_edge_function_arns
            ]

        default_cache_behavior = cloudfront.DefaultCacheBehavior(**cache_params)

        params: dict[str, Any] = {}
        if self.logging_bucket is not None:
            params["Logging"] = cloudfront.Logging(
                Bucket=self.logging_bucket,
                Prefix=self.logging_prefix if self.logging_prefix is not None else "",
                IncludeCookies=self.logging_include_cookies
                if self.logging_include_cookies is not None
                else False,
            )

        return cloudfront.Distribution(
            name_to_id(self.name),
            DistributionConfig=cloudfront.DistributionConfig(
                Aliases=self.aliases,
                DefaultRootObject=self.root_object,
                DefaultCacheBehavior=default_cache_behavior,
                Enabled="True",
                HttpVersion="http2",
                Origins=[origin],
                ViewerCertificate=cloudfront.ViewerCertificate(
                    AcmCertificateArn=self.certificate_arn,
                    SslSupportMethod="sni-only",
                    MinimumProtocolVersion="TLSv1.2_2021",
                ),
                **params,
            ),
        )

    @property
    def origin_access_identity(self) -> cloudfront.CloudFrontOriginAccessIdentity:
        """Return cloudformation access identity.

        It is needed to be used as principal for s3 bucket access policy.
        """
        if self._origin_access_identity is None:
            cf_oai_config = cloudfront.CloudFrontOriginAccessIdentityConfig(
                Comment=f"{self.name} Cloudfront origin access identity"
            )
            self._origin_access_identity = cloudfront.CloudFrontOriginAccessIdentity(
                name_to_id(f"{self.name}-cloudfront-oai"),
                CloudFrontOriginAccessIdentityConfig=cf_oai_config,
            )
        return self._origin_access_identity

    def add_cache_invalidation(self, stack: Stack) -> list[AWSObject]:
        """Return resources invalidating cache when objects are pushed to s3.

        A lambda is called at each s3 object update to invalidate cloudformation
        cache for the updated object.
        """
        lambda_name = f"{self.name}-cache-invalidation-lambda"
        lambda_policy = ManagedPolicy(
            name=f"{lambda_name}-policy",
            description=f"managed policy used by {lambda_name}",
            path=f"/{stack.name}/",
            statements=[
                Allow(
                    action=["cloudfront:CreateInvalidation"],
                    resource=Join(
                        "",
                        ["arn:aws:cloudfront::", AccountId, ":distribution ", self.id],
                    ),
                )
            ],
        )
        lambda_role = Role(
            name=f"{lambda_name}-role",
            description=f"role assumed by {lambda_name}",
            path=f"/{stack.name}/",
            trust=Trust(services=["lambda"]),
            managed_policy_arns=[lambda_policy.arn],
        )

        # Get first part of invalidation lambda code from a file
        with open(
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "data",
                "lambda_invalidate_head.py",
            )
        ) as lf:
            lambda_code = lf.read().splitlines()

        # Complete it with the part depending on the distribution id
        lambda_code.extend(
            [
                "    client.create_invalidation(",
                Sub(
                    "        DistributionId='${distribution_id}',",
                    distribution_id=self.id,
                ),
                "        InvalidationBatch={",
                "            'Paths': {'Quantity': 1, 'Items': path},",
                "            'CallerReference': str(time.time()),",
                "        },",
                "    )",
            ]
        )
        lambda_function = Function(
            name_to_id(lambda_name),
            description=(
                f"lambda invalidating cloudfront cache when "
                f"{self.bucket.name} objects are updated"
            ),
            handler="invalidate.handler",
            role=lambda_role,
            code_zipfile=Join("\n", lambda_code),
            runtime="python3.9",
        )

        sns_topic = Topic(name=f"{self.name}-invalidation-topic")
        sns_topic.add_lambda_subscription(
            function=lambda_function,
            delivery_policy={"throttlePolicy": {"maxReceivesPerSecond": 10}},
        )
        # Trigger the invalidation when a file is updated
        self.bucket.add_notification_configuration(
            event="s3:ObjectCreated:*", target=sns_topic, permission_suffix=self.name
        )

        result = [
            resource
            for construct in (lambda_policy, lambda_role, lambda_function, sns_topic)
            for resource in construct.resources(stack)
        ]
        return result

    @property
    def domain_name(self) -> GetAtt:
        """Return cloudfront distribution domain name."""
        return GetAtt(name_to_id(self.name), "DomainName")

    @property
    def id(self) -> Ref:
        """Return cloudfront distribution id."""
        return Ref(name_to_id(self.name))

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return list of AWSObject associated with the construct."""
        # Add bucket policy granting read access to te cloudfront distribution
        self.add_oai_access_to_bucket()

        result = [
            *self.bucket.resources(stack),
            self.cache_policy,
            self.distribution,
            self.origin_access_identity,
        ]

        # Add a lambda invalidating cloudfront cache when bucket objects are modified
        result.extend(self.add_cache_invalidation(stack))

        # Add route53 records if needed
        if self.r53_route_from:
            for zone_id, domain in self.r53_route_from:
                result.append(
                    route53.RecordSetType(
                        name_to_id(f"{self.name}-{domain}-r53-rset"),
                        AliasTarget=route53.AliasTarget(
                            DNSName=self.domain_name,
                            # Z2FDTNDATAQYW2 is always the hosted zone ID when you
                            # create an alias record that routes traffic to a
                            # CloudFront distribution
                            HostedZoneId="Z2FDTNDATAQYW2",
                        ),
                        Name=domain,
                        HostedZoneId=zone_id,
                        Type="A",
                    )
                )
        return result
